/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */


char hfswrap_readme[] = {
"Why can't you see your files?\r\
\r\
This hard disk is formatted with the Mac OS Extended format. Your files and \
information are still on the hard disk, but you cannot access them with the \
version of system software you are using.\r\
\r\
How can you access your files?\r\
\r\
To access your files you must mount this hard disk on a computer that has \
Mac OS 8.1 or later installed. To determine the version of system software \
youÕre currently using, choose About This Computer from the Apple menu. If \
youÕre using a version of the Mac OS earlier than 8.1, you must do one of the \
following: A) upgrade the system software on your computer, B) start up the \
computer from a hard disk or CD that has Mac OS 8.1 or later, or C) connect \
the hard disk to another computer with Mac OS 8.1 or later installed.\r\
\r\
If you want to access the files without upgrading your system software, \
start up your computer with the ÒMac OS 8.1Ó CD or the ÒDisk Tools PPCÓ \
disk, then access your files. Apple recommends that you have the ÒMac OS \
8.1Ó CD if you plan to reformat any hard disk using Mac OS Extended format.\r\
\r\
To continue to use this hard disk with this computer, you must upgrade your \
system software to Mac OS 8.1.\r\
\r\
How do you upgrade your system software?\r\
\r\
If you have a version of system software earlier than Mac OS 8, you can order \
Mac OS 8.1 on the Internet or buy it at a local Apple software reseller.\r\
\r\
If you have Mac OS 8 on your computer, you can download the Mac OS 8.1 \
update from the Internet at http://www.info.apple.com.\r\
\r\
\r\
\r\
Copyright 1998  Apple Computer, Inc. All rights reserved.\r\
Apple and Mac OS are trademarks of Apple Computer, Inc., registered in the \
U.S. and other countries. PowerPC is a registered trademark of International \
Business Machines Corporation, used under license therefrom.\r\
\0"
};

